# yaranullin/drawingarea.py

#
# Copyright (c) 2010 Marco Scopesi <marco.scopesi@gmail.com>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
import pygame
import pygame.locals as PL

import pgu_gui as gui
from board import Board
from conf import conf
from dialogs import SetBoardDialog, AddPawnDialog
from pgu_gui import pguglobals
from toolkit import load_image


class initial_value(object):

    def __init__(self, value=None):
        self.__i = value
        try:
            # Initial value
            self.p = value.copy()
        except:
            # Present value
            self.p = value

    @property
    def i(self):
        return self.__i

    @i.setter
    def i(self, value):
        if self.__i == None:
            self.__i = value
        else:
            print "Initial value already set"


class RenderingArea(gui.Widget):

    def __init__(self, width=None, height=None):
        gui.Widget.__init__(self, width=width, height=height,\
            focusable=False)
        self.game = pguglobals.app
        self.sprites = {}
        self.sprites['pawns'] = pygame.sprite.RenderUpdates()
        self.grid = False
        self.hud = True
        self.hud_font = pygame.font.Font(None, 24)

    def set_board(self, img_name, real_w):
        self.real_w = real_w
        self.img_name = img_name

        w, h = self.style.width, self.style.height
        self.view = pygame.rect.Rect(0, 0, w, h)

        img, rect = load_image(img_name)
        self.img = initial_value(img)
        self.img_rect = initial_value(rect)
        # Tiles are always 1.5 meters wide, so their number is:
        real_tile_w = 1.5
        nx = int(self.real_w / real_tile_w)
        # So, the width of the tiles in pixels is:
        back_w = self.img_rect.i.width
        self.tile_w = initial_value(int(back_w / nx))
        # Finally, the number of tiles in the y direction is:
        back_h = self.img_rect.i.height
        ny = int(back_h / self.tile_w.i)

        self.board = Board((nx, ny), self.game, self)

        #Connect callbacks
        self.connect(PL.MOUSEBUTTONDOWN, self.mousebuttondown)
        self.connect(PL.MOUSEBUTTONUP, self.mousebuttonup)
        self.connect(PL.MOUSEMOTION, self.mousemotion)
        self.connect(PL.KEYDOWN, self.keydown)

    def get_rect_at_pos(self, pos):
        tw = self.tile_w.p
        dx, dy = self.view.left, self.view.top
        x = tw * pos[0] - dx
        y = tw * pos[1] - dy
        return pygame.rect.Rect(x, y, tw, tw)

    def get_pos_at_point(self, point):
        tw = self.tile_w.p
        dx, dy = self.view.left, self.view.top
        x = point[0] + dx
        y = point[1] + dy
        pos = int(x / tw), int(y / tw)
        return pos

    def set_board_from_dialog(self, value):
        board_name, real_w = value['fname'].value, int(value['real_w'].value)
        self.game.set_board(board_name, real_w)

    def add_pawn_from_dialog(self, value):
        pname = value['pname'].value
        fname = value['fname'].value
        initiative = int(value['initiative'].value)
        pos = self.get_pos_at_point(self.click_button3)
        self.game.add_pawn(pname, initiative, pos, fname)

    # Events related funcions
    def event(self, e):
        used = False
        if ((e.type == PL.KEYDOWN) and (e.key == PL.K_b)):
            dlg = SetBoardDialog()
            dlg.connect(gui.CHANGE, self.set_board_from_dialog, dlg.value)
            dlg.connect(gui.CHANGE, dlg.close, None)
            dlg.open()
            used = True
        if (e.type in self.connects):
            used = True
        return used

    def mousebuttonup(self, _event):
        # Using _clicked we know if the release of the button takes place
        # after a drag or it is a simple click
        e = _event
        if ((e.button == 1) and (self._clicked == True)):
            self._clicked = False
            pos = self.get_pos_at_point(pygame.mouse.get_pos())
            return self.board.move_pawn(pos)
        else:
            return False

    def mousebuttondown(self, _event):
        e = _event
        if e.button == 1:
            self._clicked = True
            return True
        elif e.button == 3:
            if hasattr(self, 'board'):
                self.click_button3 = pygame.mouse.get_pos()
                x, y = self.get_pos_at_point(self.click_button3)
                if self.board.grid[x][y] == []:
                    dlg = AddPawnDialog()
                    dlg.connect(gui.CHANGE, self.add_pawn_from_dialog,\
                        dlg.value)
                    dlg.connect(gui.CHANGE, dlg.close, None)
                    dlg.open()
                else:
                    pawns_to_remove = self.board.grid[x][y]
                    self.game.remove_pawns(pawns_to_remove)
                    self.repaint()
                return True
        elif e.button == 4:
            return self.zoom('in')
        elif e.button == 5:
            return self.zoom('out')

    def keydown(self, _event):
        e = _event
        if e.key == PL.K_g:
            return self.togglegrid()
        elif e.key == PL.K_o:
            return self.zoom('orig')
        elif e.key == PL.K_h:
            return self.togglehud()
        elif e.key == PL.K_SPACE:
            self.game.next_pawn()
            self.repaint()
            return True

    def mousemotion(self, _event):
        e = _event
        self._clicked = False
        if pygame.mouse.get_pressed()[0] == True:
            return self.scroll(e.rel)

    def scroll(self, rel):
        dx, dy = - rel[0], - rel[1]
        img_rect = self.img_rect.p
        view = self.view
        view.top += dy
        view.left += dx
        view.bottom = min(view.bottom, img_rect.bottom)
        view.right = min(view.right, img_rect.right)
        view.top = max(view.top, img_rect.top)
        view.left = max(view.left, img_rect.left)
        self.view = view
        self.repaint()
        return True

    def zoom(self, direction):
        if direction == 'orig':
            self.tile_w.p = self.tile_w.i
            self.img.p, self.img_rect.p =\
                self.img.i.copy(), self.img_rect.i.copy()
            self.repaint()
            return True
        if direction == 'in':
            self.tile_w.p += 2
            if self.tile_w.p > self.tile_w.i * 2:
                self.tile_w.p = self.tile_w.i * 2
                return True
        elif direction == 'out':
            self.tile_w.p -= 2
            if self.tile_w.p < int(self.tile_w.i / 2):
                self.tile_w.p = int(self.tile_w.i / 2)
                return True
        nx, ny = self.board.size
        new_w = self.tile_w.p * nx
        new_h = self.tile_w.p * ny
        self.img.p = pygame.transform.smoothscale(self.img.i,\
            (new_w, new_h))
        self.img_rect.p = self.img.p.get_rect()
        self.repaint()
        return True

    def paint(self, surf):
        surf.fill((0, 0, 0))
        if hasattr(self, 'board'):
            view = self.view.clip(self.img_rect.p)
            back_img = self.img.p.subsurface(view).copy()
            if self.grid:
                self.paintgrid(back_img, view)
            self.sprites['pawns'].update(self)
            self.sprites['pawns'].draw(back_img)
            if self.hud:
                self.painthud(back_img)
            surf.blit(back_img, (0, 0))

    def update(self, surf):
        if hasattr(self, 'board'):
            view = self.view.clip(self.img_rect.p)
            back_img = self.img.p.subsurface(view).copy()
            if self.grid:
                self.paintgrid(back_img, view)
            self.sprites['pawns'].clear(surf, back_img)
            self.sprites['pawns'].update(self)
            rects = self.sprites['pawns'].draw(surf)
            if self.hud:
                self.painthud(surf)
            return rects

    def paintgrid(self, surf, view):
        grid_surf = \
            pygame.surface.Surface((self.img_rect.p.width,\
            self.img_rect.p.height), PL.SRCALPHA).convert_alpha()
        fill_col = pygame.color.Color(conf['colors-grid_fill'])
        border_col = pygame.color.Color(conf['colors-grid_border'])
        grid_surf.fill(fill_col)
        nx, ny = self.board.size
        for y in range(ny):
            pygame.draw.line(grid_surf, border_col, (0, y * self.tile_w.p),\
                (self.img_rect.p.width, y * self.tile_w.p))
        for x in range(nx):
            pygame.draw.line(grid_surf, border_col, (x * self.tile_w.p, 0),\
                (x * self.tile_w.p, self.img_rect.p.height))
        surf.blit(grid_surf.subsurface(view), (0, 0))

    def painthud(self, s):
        fnt = self.hud_font
        surfs = []
        color = pygame.color.Color(conf['colors-hud_pawns'])
        color_active_pawn = pygame.color.Color(conf['colors-hud_current_pawn'])
        for pawn in self.board.initiatives:
            if pawn == self.board.current_pawn:
                surfs.append(fnt.render(pawn.name, True, color_active_pawn))
            else:
                surfs.append(fnt.render(pawn.name, True, color))
        next_y = 2
        x = 5
        for surf in surfs:
            rect = surf.get_rect(top=next_y, left=x)
            s.blit(surf, rect)
            next_y = rect.bottom + 2

    def togglegrid(self):
        self.grid = not self.grid
        self.repaint()
        return True

    def togglehud(self):
        self.hud = not self.hud
        self.repaint()
        return True
